# encoding: utf-8
from __future__ import unicode_literals

from django.db.models import Sum, Q

from kinder.core.group.models import (
    Pupil,
    Group,
    GroupStatusEnum,
)
from kinder.core.direct.models import Direct, DirectStatusSewn
from kinder.core.declaration.enum import DeclarationSourceEnum
from kinder.core.declaration.models import Declaration, DeclarationStatusLog
from kinder.core.declaration_status.enum import DSS
from kinder.core.children.models import Delegate, Children

from edu_to_analytics.service.utils import PERIOD_TODAY, PERIOD_PERIOD

from .interfaces import CalculatedIndex, ModelIndex
from .logs import (
    ParentsLog,
    ParentsSnilsLog,
    DirectTemporaryLog,
    DeclarationWorkTypeLog,
)
from .period import period


DEFAULT_PERIOD = period(PERIOD_TODAY)
#: Заявление с такими статусами считается учавствующим в очереди
QUEUE_STATUS_CODES = (
    DSS.WANT_CHANGE_DOU,
    DSS.REGISTERED,
    DSS.PRIV_CONFIRMATING,
)
#: Статусы при которых направления считаются выданными
ISSUED_DIRECT_STATUS_CODES = (
    DirectStatusSewn.NEW,
    DirectStatusSewn.CONFIRM,
)


class NumberPupil(ModelIndex):
    """
    Количество воспитанников на текущую дату.

    Рассчитывается только на текущую дату
    Воспитанники считаются только в фактических группах
    """
    model = Pupil
    extra_filter = Q(grup__status=GroupStatusEnum.FACT)


class NumberPlaceKNDG(ModelIndex):
    """
    Количество мест в детском саду.

    По всем фактическим группам суммируется максимальная наполняемость
    """
    model = Group
    extra_filter = Q(status=GroupStatusEnum.FACT)

    def calculate(self, unit, period):
        total = (
            self.qs(unit).filter(self.extra_filter)
            .aggregate(total=Sum('max_count'))['total'])
        # Aggregate может вернуть None, и далее по коду
        # при попытке вычесть из None будет боль
        return total or 0


class NumberFreePlaceKNDG(CalculatedIndex):
    """
    Количество свободных мест.

    Показатель "Количество мест в детском саду" минус показатель
    "Количество воспитанников на текущую дату"
    """
    def calculate(self, unit, period):
        total = (
            NumberPlaceKNDG().calculate(unit, DEFAULT_PERIOD) -
            NumberPupil().calculate(unit, DEFAULT_PERIOD)
        )
        return total


class NumberChildQueue(ModelIndex):
    """Количество детей в очереди по приоритету ДОО."""
    model = Children
    extra_filter = Q(declaration__status__code__in=QUEUE_STATUS_CODES)
    # extra_param = номер приоритета

    def calculate(self, unit, period):
        return (
            self.qs(unit)
            .filter(self.extra_filter)
            .filter(declaration__declarationunit__ord=self.extra_param)
            .count())


class NumberChildQueueSource(ModelIndex):
    """Количество детей в очереди по Источнику заявки."""
    model = Children
    extra_filter = Q(declaration__status__code__in=QUEUE_STATUS_CODES)
    # extra_param = id источника заявки

    def calculate(self, unit, period):
        source_id = int(getattr(
            DeclarationSourceEnum, self.extra_param))
        return (
            self.qs(unit)
            .filter(self.extra_filter)
            .filter(declaration__source=source_id)
            .count())


class NumberChildWantChange(ModelIndex):
    """Количество детей желающих быть перераспределенными.

    Сделано по-другому, т.к. для заявлений есть специальная модель,
    отслеживающая изменения - DeclarationStatusLog, да и запросами быстрее.
    """
    model = Declaration

    def calculate(self, unit, period):
        unit_declarations = self.qs(unit)

        current_declarations_count = unit_declarations.filter(
                status__code=DSS.WANT_CHANGE_DOU
            ).count()

        if period.type == PERIOD_TODAY:
            # просто берем все число заявок в статусе "Желает изменить ДОУ"
            # по учреждению на текущий момент
            return current_declarations_count
        else:
            # считаем переходы заявок в статус
            # (для каждой заявки берем переход один раз)
            status_change_count = DeclarationStatusLog.objects.filter(
                declaration__in=unit_declarations,
                datetime__lte=period.end,
                datetime__gte=period.start,
                status__code=DSS.WANT_CHANGE_DOU
            ).order_by(
                'declaration__id'
            ).distinct(
                'declaration__id'
            ).count()

            if period.type == PERIOD_PERIOD:
                # возвращаем изменения за период
                return status_change_count
            else:
                # возвращаем разницу от текущего показателя
                return current_declarations_count - status_change_count


class NumberParents(ModelIndex):
    """Количество внесенных в систему родителей."""
    model = Delegate

    logs = [ParentsLog]


class NumberParentsSNILS(ModelIndex):
    """Количество внесенных в систему родителей со СНИЛС."""
    model = Delegate
    extra_filter = Q(snils__isnull=False) & ~Q(snils='')

    logs = [ParentsSnilsLog]


class NumnerChildQueueShortTerm(ModelIndex):
    """Кол-во желающих посещать ДОУ в режиме кратковременного пребывания."""

    model = Children
    field = 'declaration'
    extra_filter = (
        Q(declaration__consent_short_time_group=True) &
        Q(declaration__status__code=DSS.REGISTERED)
    )

    logs = [DeclarationWorkTypeLog]


class NumberFreePlaceDirectionWithout(CalculatedIndex):
    """Количество свободных мест без учёта уже выданных направлений."""

    def calculate(self, unit, period):
        issued_directs_count = Direct.objects.filter(
            group__unit_id=unit['id'],
            status__code__in=ISSUED_DIRECT_STATUS_CODES).count()
        free_places = NumberFreePlaceKNDG().calculate(unit, DEFAULT_PERIOD)
        return issued_directs_count - free_places


class NumberDirection(ModelIndex):
    """Количество выданных направлений."""
    model = Direct
    extra_filter = Q(group__status=GroupStatusEnum.PLAN) & Q(temporary=False)

    logs = [DirectTemporaryLog]
